#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include "fsactions.h"
#include "fusestructs.h"

/*  Things to set for directory:
 *      Name, isValid, inode, isDirectory=1
 */

/*  Things to set for file:
 *      Name, size, isValid, inode, isDirectory=0
 */

// Adds a file to the filesystem
// TODO: Fix implementation to add starting at Inode 0
void AddFileToFS(FileSystem* fs, char* fname)
{
    struct stat tempstat;
    ino_t currentInode = 0;
    ino_t newInode;
    int newDirectBlock;
    int tempfbl[4];
    char* tempname;
    stat(fname, &tempstat);
    // Get first directory, labelled by 0 to splitLocation-1,
    //      assuming splitLocation not 0;
    int splitLocation = FindNextDirectory(fname);
    if (splitLocation == 0)
    {
        fname = fname + 1; 
        splitLocation = FindNextDirectory(fname);
    }
    // While there is a '/', make a new inode out of left and keep right side
    while (splitLocation > -1)
    {
        // First: check if directory already exists in current inode
        //          if yes, then go to that inode and start with next directory
        // Second: get a number for the next free direct block
        // Third: get a free inode number
        // Fourth: set info for direct block
        tempname = strndup(fname, splitLocation);
        if (!IsDirectorySetup(fs->inodes[currentInode], &currentInode, tempname))
        {
            for (int i = 0; i < 4; i++)
                tempfbl[i] = (currentInode*4) + i;
            newDirectBlock = GetFreeBlockNumber(tempfbl, 4);
            newInode = GetFreeInodeNumber(fs->inodes);
            printf("Name held: |%s|\n", tempname);
            printf("Block: |%i|\n", newDirectBlock);
            strcpy(fs->inodes[currentInode].blocks[newDirectBlock].name, tempname);
            free(tempname);
            fs->inodes[currentInode].blocks[newDirectBlock].isValid = 1;
            fs->inodes[currentInode].blocks[newDirectBlock].isDirectory = 1;
            fs->inodes[currentInode].blocks[newDirectBlock].inode = newInode;
            currentInode = newInode;
            fs->inodes[currentInode].isValid = 1;
        }
        fname = fname + splitLocation + 1;
        splitLocation = FindNextDirectory(fname);
    }
    for (int i = 0; i < 4; i++)
        tempfbl[i] = (currentInode*4) + i;
    newDirectBlock = GetFreeBlockNumber(tempfbl, 4);
    fs->inodes[currentInode].blocks[newDirectBlock].isValid = 1;
    fs->inodes[currentInode].blocks[newDirectBlock].size = tempstat.st_size;
    strcpy(fs->inodes[currentInode].blocks[newDirectBlock].name, fname);
    return;
}

// List path names of all used inodes
void ListFS(FileSystem* fs)
{
    Inode* fsRoot = &(fs->inodes[0]);
    printf("/\n");
    for (int i = 0; i < fs->superBlock.inodeCount; i++)
    {
        if (!(fsRoot->blocks[i].isValid))
            continue;
        printf("%s\n", fsRoot->blocks[i].name);
        if (fsRoot->blocks[i].isDirectory)
            _ListFS_(fs, fsRoot->blocks[i].inode, 1);
    }
    return;
}

void _ListFS_(FileSystem* fs, ino_t inodeNumber, short unsigned numtabs)
{
    for (int i = 0; i < fs->superBlock.inodeCount; i++)
    {
        if (!(fs->inodes[inodeNumber].blocks[i].isValid))
            continue;
        for (int j = 0; j < numtabs; j++)
            printf("\t");
        printf("%s\n", fs->inodes[inodeNumber].blocks[i].name);
        if (fs->inodes[inodeNumber].blocks[i].isDirectory)
            _ListFS_(fs, fs->inodes[inodeNumber].blocks[i].inode, numtabs + 1);
    }
}

// Remove a file from the filesystem
// TODO: Fix function so that it adds directories, then the file after
void RemoveFileFromFS(FileSystem* fs, char* fname)
{

}

// Extract a file from the filesystem
// TODO: Fix function so that it adds directories, then the file after
void ExtractFileFromFS(FileSystem* fs, char* fname)
{
    
}


// Get the next free inode number
ino_t GetFreeInodeNumber(Inode inodes[])
{
    for (int i = 0; i < DEFAULTINODEMAX; i++)
        if (!(inodes[i].isValid))
            return i;
    return -1;
}

// Gets next free inode number and returns it
ino_t GetFreeBlockNumber(int fbl[], unsigned short size)
{
    int bitPosition;
    for (int i = 0; i < size; i++)
    {
        bitPosition = FindEmptyBitPosition(fbl[i]);
        if (bitPosition >= 0)
        {
            fbl[i] = (fbl[i] | (1 << bitPosition));
            return ((i*32) + bitPosition);
        }
    }
    return -1;
}

// Finds next empty bit in Free Block List
int FindEmptyBitPosition(int number)
{
    for (int i = 0; i < 32; i++)
        if ((~number & (1 << i)) == (1 << i))
            return i;
    return -1;
}

// Finds the next split character in the path
int FindNextDirectory(char* fname)
{
    int i = 0;
    while (fname[i] != '\0')
    {
        if (fname[i] == '/')
            return i;
        i++;
    }
    return -1;
}

// Bool function checks if directory is setup. If it is, then sets the inode
int IsDirectorySetup(Inode inode, ino_t* inodePosition, char* directoryName)
{
    for (int i = 0; i < 128; i++)
    {
        if (!(inode.blocks[i].isValid))
            continue;
        if (!(inode.blocks[i].isDirectory))
            continue;
        if (strcmp(inode.blocks[i].name, directoryName) == 0)
        {
            *inodePosition = i;
            return 1;
        }
    }
    return 0;
}

// Function for setting all defaults of filesystem
void SetFileSystemDefaults(FileSystem* fs)
{
    fs->superBlock.blockSize = DEFAULTBLOCKSIZE;
    fs->superBlock.blockCount = DEFAULTBLOCKMAX;
    fs->superBlock.inodeCount = DEFAULTINODEMAX;
    for (int i = 0; i < 4; i++)
        fs->fbl.freeList[i] = 0;
    for (int i = 0; i < DEFAULTINODEMAX; i++)
        fs->inodes[i].isValid = 0;
    return;
}

// Sets first inode as root directory
void SetupRootDirectory(FileSystem* fs)
{
    fs->inodes[0].isValid = 1;
    return;
}
